/*
 * Copyright (C) 2002,2003 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: world.cc,v 1.72 2003/09/12 21:46:25 dheck Exp $
 */
#include "world.hh"
#include "objects.hh"
#include "actors.hh"
#include "items.hh"
#include "laser.hh"
#include "stones.hh"
#include "display.hh"
#include "player.hh"
#include "sound.hh"
#include "px/px.hh"
#include "px/dict.hh"
#include "lua.hh"
#include "options.hh"
#include "game.hh"

#include <iostream>
#include <algorithm>
#include <functional>
#include <numeric>
#include <cassert>

// remove comment from define below to switch on verbose messaging
// note: VERBOSE_MESSAGES is defined in multiple source files!
// #define VERBOSE_MESSAGES

using namespace std;
using namespace world;
using namespace px;

namespace
{
    class Field;

    typedef px::Array2<Field> FieldArray;
    typedef std::vector<Actor*> ActorList;

    typedef vector<ForceField*> ForceList;
    typedef vector<StoneContact> StoneContactList;
}

namespace
{
    Object *GetObject (const GridLoc &l)
    {
        switch (l.layer) {
        case GRID_FLOOR:  return GetFloor(l.pos);
        case GRID_ITEMS:  return GetItem(l.pos);
        case GRID_STONES: return GetStone(l.pos);
        default: return 0;
        }
    }

    class Signal {
    public:
        virtual ~Signal() {};
        virtual void emit_from(Object *source, int value) = 0;
        virtual Object *get_target() = 0;
        virtual Object *get_source() = 0;
    };

    class SimpleSignal : public Signal {
    public:
        SimpleSignal (Object *src, GridLoc dstloc, const string &msg)
        : source (src), destloc(dstloc), message(msg)
        {}

        Object *get_target() {
            return GetObject(destloc);
        }
        Object *get_source() {
            return source;
        }

        void emit_from (Object *src, int value) {
            if (source == src) {
                Object *dst = get_target();

#if defined(VERBOSE_MESSAGES)
                src->warning("emit_from: msg='%s' dest=%i/%i obj=%p",
                             message.c_str(),
                             destloc.pos.x, destloc.pos.y,
                             dst);
#endif // VERBOSE_MESSAGES

                SendMessage (dst, message, value);
            }
#if defined(VERBOSE_MESSAGES)
//             else {
//                 fprintf(stderr,
//                         "emit_from source object does not match -- signal '%s' dropped (expected=%p=%s found=%p=%s)\n",
//                         message.c_str(), source, source->get_kind(), src, src->get_kind());
//             }
#endif // VERBOSE_MESSAGES
        }
    private:
        Object *source;
        GridLoc destloc;
        string message;
    };

    class SignalList {
        typedef vector<Signal *> ListT;
        ListT m_signals;

    public:
        ~SignalList() { clear(); }
        void clear() {
            delete_sequence (m_signals.begin(), m_signals.end());
            m_signals.clear();
        }

        void add (Signal *sig) { m_signals.push_back(sig); }

        void emit_from (Object *source, int value) {
            // signals may have side effects. To minimize them
            //   1. collect all targets and then
            //   2. emit signals to targets

            size_t          size = m_signals.size();
            vector<Object*> targets;
            targets.resize(size);

            for (unsigned i=0; i<size; ++i) {
                targets[i] = m_signals[i]->get_target();
            }

            for (unsigned i=0; i<size; ++i) {
                if (m_signals[i]->get_target() == targets[i])
                    m_signals[i]->emit_from (source, value);
            }
        }

        Object *find_single_destination(Object *src) {
            Object *found = 0;
            size_t size = m_signals.size();

            for (unsigned i = 0; i<size; ++i) {
                if (m_signals[i]->get_source() == src) {
                    Object *candidate = m_signals[i]->get_target();
                    if (candidate) {
                        if (!found)
                            found = candidate;
                        else if (candidate != found) {
                            found = 0; // multiple targets
                            break;
                        }
                    }
                }
            }

            return found;
        }
    };
}

//----------------------------------------
// Rubber band
//----------------------------------------
namespace world
{
    class RubberBand {
    public:
        RubberBand (Actor *a1, Actor *a2, double strength=10, double length=2);
        RubberBand (Actor *a1, Stone *st, double strength=10, double length=2);
        ~RubberBand();

        void tick(double);
        V2 get_force(Actor *a);

        Actor *get_actor() const { return actor; }
        Actor *get_actor2() const { return actor2; }
        Stone *get_stone() const { return stone; }

	double get_strength() const { return strength; }
	double get_length() const { return length; }
    private:
        V2 get_p1() const;
        V2 get_p2() const;

        // Variables.
        Actor *actor, *actor2;
        Stone *stone;
        display::RubberHandle model;
        double strength, length;
    };
}

RubberBand::RubberBand (Actor *a1, Actor *a2, double strength_, double length_)
: actor(a1), actor2(a2), stone(0),
  model(display::AddRubber(get_p1(),get_p2())),
  strength(strength_), length(length_)
{
    assert(actor);
    assert(length >= 0);
}

RubberBand::RubberBand (Actor *a1, Stone *st, double strength_, double length_)
: actor(a1), actor2(0), stone(st), model(0),
  strength(strength_), length(length_)
{
    assert(actor);
    assert(length >= 0);
    model = display::AddRubber(get_p1(), get_p2());
}

RubberBand::~RubberBand() {
    model.kill();
}

void RubberBand::tick(double)
{
    V2 v = get_p2()-get_p1();
    double vv = px::length(v);

    if (vv > length) {
        V2 force = v * strength*(vv-length)/vv;
        V2 f(force[0],force[1]);
        actor->add_force(f);
        if (actor2)
            actor2->add_force(-f);
    }

    model.update_first (get_p1());
    model.update_second (get_p2());
}

V2 RubberBand::get_force(Actor */*a*/)
{
    return V2();
}

V2 RubberBand::get_p1() const
{
    return V2(actor->get_pos()[0], actor->get_pos()[1]);
}

V2 RubberBand::get_p2() const
{
    if (!stone)
        return V2(actor2->get_pos()[0], actor2->get_pos()[1]);
    else
        return stone->get_pos().center();
}

//----------------------------------------
// Field
//----------------------------------------
namespace
{
    class Field {
    public:
	Field();
	~Field();

        Floor *floor;
        Item *item;
        Stone *stone;
    };
}


Field::Field()
{
    floor=0;
    item=0;
    stone=0;
}

Field::~Field()
{
    DisposeObject(floor);
    DisposeObject(item);
    DisposeObject(stone);
}

/*
** This class implements the "force field" that accelerates objects
** when the mouse is moved.  Only objects that have the "mouseforce"
** attribute set are affected by this force field.
*/
namespace
{
    class MouseForce {
    public:
        void set_force(V2 f) { force=f; }
        void add_force(V2 f) { force+=f; }

        V2 get_force(Actor *a) {
            double mouseforce = 0;
            a->double_attrib("mouseforce", &mouseforce);
            if (a->is_flying() || a->is_dead())
                return V2();
            else
                return force * mouseforce;
        }

        void tick(double /*dtime*/) {
            force=V2();
        }
    private:
        V2 force;
    };
}

StoneContact::StoneContact(Actor *a, GridPos p,
                           const V2 &cp, const V2 &n)
: actor(a), stonepos(p),
  response(STONE_PASS),
  contact_point(cp),
  normal(n),
  is_collision(false),
  fake_collision(false),
  new_collision(false),
  is_contact(true)
{}

StoneContact::StoneContact()
: is_collision(false),
  fake_collision(false),
  new_collision(false),
  is_contact(false)
{}

DirectionBits
world::contact_faces(const StoneContact &sc)
{
    using namespace enigma;

    int dirs = NODIRBIT;

    if (sc.normal[0] < 0)
        dirs |= WESTBIT;
    else if (sc.normal[0] > 0)
        dirs |= EASTBIT;
    if (sc.normal[1] < 0)
        dirs |= NORTHBIT;
    else if (sc.normal[1] > 0)
        dirs |= SOUTHBIT;

    return DirectionBits(dirs);
}

Direction
world::contact_face(const StoneContact &sc)
{
    using namespace enigma;
    if (sc.normal == V2(-1,0))
        return WEST;
    else if (sc.normal == V2(1,0))
        return EAST;
    else if (sc.normal == V2(0,-1))
        return NORTH;
    else if (sc.normal == V2(0,1))
        return SOUTH;
    else
        return NODIR;
}

// ------------------
//      Scrambles
// ------------------
// Stores all positions plus direction where puzzle scrambling should be done.

namespace
{
    struct Scramble {
        GridPos   pos;
        Direction dir;
        int       intensity;

        Scramble(GridPos p, Direction d, int i) : pos(p), dir(d), intensity(i) {}

        bool expired() const { return intensity<1; }
    };
}



//======================================================================
// VARIABLES
//======================================================================

namespace
{
    class World {
    public:
        virtual ~World() {}
        virtual void name_object (Object *obj, const string &name) = 0;
    };

    class WorldImpl : public World {
    public:
        WorldImpl(int ww, int hh);
        ~WorldImpl();

        bool contains (GridPos p) {
            return (p.x>=0 && p.y>=0 && p.x<w && p.y<h);
        }

        bool is_border (GridPos p) {
            return(p.x==0 || p.y==0 || p.x==w-1 || p.y==h-1);
        }

        Field *get_field (GridPos p) {
            if (this->contains(p))
                return &fields(p.x,p.y);
            else
                return 0;
        }

        void    name_object (Object *obj, const string &name);
        void    unname (Object *);
        Object *get_named (const string &);

        void tick (double dtime);
        void remove (ForceField *ff);

        void add_scramble(GridPos p, Direction dir);
        void scramble_puzzles();

    public:
        // Variables.
        FieldArray           fields; // Contains floors, items, etc.
        int                  w, h; // Width and height of the level
        ForceList            forces;
        ActorList            actorlist; // List of movable, dynamic objects
        vector<RubberBand *> rubbers;
        SignalList           signals;
        MouseForce           mouseforce;
        int                  scrambleIntensity;

    private:
        /* Objects can be named with NameObject and referenced with
           GetNamedObject.  This dictionary maps the names to the
           actual objects. */
        px::Dict<Object *> m_objnames;

        list<Scramble> scrambles;
    };

    auto_ptr<WorldImpl> level;
}


WorldImpl::WorldImpl(int ww, int hh) : fields(ww,hh)
{
    w = ww;
    h = hh;

    scrambleIntensity = options::Difficulty == DIFFICULTY_EASY ? 3 : 10;
}

WorldImpl::~WorldImpl()
{
    fields = FieldArray(0,0);
    for_each(actorlist.begin(), actorlist.end(), mem_fun(&Actor::dispose));
    delete_sequence (rubbers.begin(), rubbers.end());
}

void WorldImpl::remove (ForceField *ff)
{
    ForceList::iterator i=find (forces.begin(), forces.end(), ff);
    if (i != forces.end())
        forces.erase(i);
}

Object *WorldImpl::get_named (const string &name)
{
    px::Dict<Object*>::iterator found = m_objnames.find(name);
    return (found != m_objnames.end()) ? found->second : 0;
}

void WorldImpl::name_object (Object *obj, const std::string &name)
{
    m_objnames.insert(name, obj); // [name] = obj;
    obj->set_attrib("name", name);
}

void WorldImpl::unname (Object *obj)
{
    assert(obj);
    string name;
    if (obj->string_attrib("name", &name)) {
        m_objnames.remove(name);
        obj->set_attrib("name", "");
    }
}

void WorldImpl::tick (double dtime)
{
    mouseforce.tick (dtime);
    for_each (forces.begin(), forces.end(),
              bind2nd(mem_fun(&ForceField::tick), dtime));

    for_each (rubbers.begin(), rubbers.end(),
              bind2nd(mem_fun(&RubberBand::tick), dtime));
}

enigma::Timer world::g_timer;

//======================================================================
//      PUZZLE SCRAMBLING
//======================================================================

void WorldImpl::add_scramble(GridPos p, Direction dir)
{
    // fprintf(stderr, "scrambles.size()=%i ; add_scramble(%i, %i, %s)\n", scrambles.size(), p.x, p.y, to_suffix(dir).c_str());
    scrambles.push_back(Scramble(p, dir, scrambleIntensity));
}

void WorldImpl::scramble_puzzles()
{
    // fprintf(stderr, "scramble_puzzles: scrambles.size()=%i\n", scrambles.size());
    while (!scrambles.empty()) {
        list<Scramble>::iterator i = scrambles.begin();
        list<Scramble>::iterator e = scrambles.end();

        for (; i != e; ++i) {
            Stone *puzz = GetStone(i->pos);
            if (puzz && i->intensity) {
                // fprintf(stderr, "Sending 'scramble' to stone at pos %i/%i\n", i->pos.x, i->pos.y);
                SendMessage(puzz, "scramble", Value(double(i->dir)));
                --i->intensity;
            }
            else {
                fprintf(stderr, "no stone found for scramble at %i/%i\n", i->pos.x, i->pos.y);
                i->intensity = 0;
            }
        }

        scrambles.remove_if(mem_fun_ref(&Scramble::expired));
    }
}


//======================================================================
// PHYSICS SIMULATION
//======================================================================

/* Calculate the total acceleration on an actor A at time TIME.  The
   actor's current position X and velocity V are also passed.  Note
   that the position and velocity entries in ActorInfo will be updated
   only after a *successful* time step, so they cannot be used
   here. */
static V2
get_accel (Actor *a, const V2 & x, const V2 & v, double time)
{
    V2 f;
    GridPos p((int)x[0], (int)x[1]);

    if (a->is_on_floor()) {
        if (Floor *floor = GetFloor(p)) {
            double friction;

            // Mouse force
            if (a->has_spikes()) {
                f += level->mouseforce.get_force(a) * 1.0;
                friction = 1.0;
            }
            else {
                f += floor->process_mouseforce(a, level->mouseforce.get_force(a));
                friction = floor->friction();
            }

            // Friction
            double vv=length(v);
            if (vv > 0) {
                V2 frictionf = v * enigma::FrictionFactor*friction;
                frictionf /= vv;
                frictionf *= pow(vv, 0.8);
                f -= frictionf;
            }

            f += floor->get_force(a);
        }

        if (Item *item = GetItem(p)) {
            f += item->get_force(a);
        }
    }

    // Electrostatic forces between actors.
    if (double q = a->get_charge()) {
        for (ActorList::iterator i=level->actorlist.begin();
             i != level->actorlist.end(); ++i)
        {
            Actor *a2 = *i;
            if (a2 == a) continue;
            if (double q2 = a2->get_charge()) {
                V2 distv = a->get_pos() - a2->get_pos();
                if (double dist = distv.normalize())
//                    f += enigma::ElectricForce * q * q2 / (dist*dist) * distv;
                    f += enigma::ElectricForce * q * q2 / (dist) * distv;
            }
        }
    }

    // All other force fields.
    for (ForceList::iterator i=level->forces.begin();
         i != level->forces.end(); ++i)
    {
        f += (*i)->get_force(a, x, v, time);
    }

    ActorInfo *ai = a->get_actorinfo();
    f += ai->force;

    return f / ai->mass;
}


/* This function performs one step in the numerical integration of an
   actor's equation of motion.  TIME ist the current absolute time and
   H the size of the integration step. */
static void
advance_actor(Actor *a, double time, double h)
{
    const double MAXVEL = 40;

    ActorInfo &ai = *a->get_actorinfo();
    V2 accel = get_accel(a, ai.pos, ai.vel, time);

    ai.vel += h*accel;
    ai.pos += h*ai.vel;

    // Limit maximum velocity
    double q = length(ai.vel) / MAXVEL;
    if (q > 1)
        ai.vel /= q;
}

/* If the actor and the stone are not in contact, `contact' is filled
   with information about the closest feature on the stone and
   `is_contact' is set to false. */
static void
find_contact_with_stone (Actor *a, StoneContact &contact, int x, int y)
{
    contact.actor      = a;
    contact.stonepos   = GridPos(x, y);
    contact.is_contact = false;

    Stone *stone = world::GetStone(contact.stonepos);
    if (!stone)
        return;

    const ActorInfo &ai = *a->get_actorinfo();
    double r = ai.radius;
    double ax = ai.pos[0];
    double ay = ai.pos[1];
    const double contact_e = 0.01;
    const double erad = 2.0/32; // edge radius

    // Closest feature == north or south face of the stone?
    if (ax>x+erad && ax<x+1-erad) {
        double dist = r+5;

        // south
        if (ay>y+1) {
            contact.contact_point = V2(ax, y+1);
            contact.normal        = V2(0,+1);
            dist                  = ay-(y+1);
        }
        // north
        else if (ay<y) {
            contact.contact_point = V2(ax, y);
            contact.normal        = V2(0,-1);
            dist                  = y-ay;
        }
	contact.is_contact = (dist-r < contact_e);
    }
    // Closest feature == west or east face of the stone?
    else if (ay>y+erad && ay<y+1-erad) {
        double dist=r+5;
        if (ax>x+1) { // east
            contact.contact_point = V2(x+1,ay);
            contact.normal        = V2(+1,0);
            dist                  = ax-(x+1);
        }
        else if (ax<x) { // west
            contact.contact_point = V2(x,ay);
            contact.normal        = V2(-1,0);
            dist                  = x-ax;
        }
	contact.is_contact = (dist-r < contact_e);
    }
    // Closest feature == any of the four corners
    else {
        int xcorner=(ax >= x+1-erad);
        int ycorner=(ay >= y+1-erad);
        double cx[2] = {erad, -erad};

        V2 corner(x+xcorner+cx[xcorner], y+ycorner+cx[ycorner]);
        V2 b=V2(ax,ay) - corner;

        contact.is_contact    = (length(b)-r-erad < contact_e);
        contact.normal        = normalize(b);
        contact.contact_point = corner + contact.normal*erad;
    }

    contact.is_collision  = contact.normal*ai.vel < 0;
    contact.response = stone->collision_response(contact);
    contact.sound = stone->collision_sound();
}

/* Enigma's stones have rounded corners; this leads to realistic
   behaviour when a marble rebounds from the corner of a single stone.
   But it's annoying when there are two adjacent stones and a marble
   rebounds from any one (or even both) corners, because it changes
   the direction of the marble quite unpredictably.

   This function modifies the contact information for two adjacent
   stones in such a way that the two collisions are treated as a
   single collision with a flat surface.
 */
static void
maybe_join_contacts (StoneContact & a, StoneContact &b)
{
//    double maxd = 4.0/32;   // Max distance between joinable collisions

    if ((a.is_contact || b.is_contact)  &&
	a.is_collision && b.is_collision &&
        a.response==STONE_REBOUND && b.response==STONE_REBOUND) // &&
//        length(a.contact_point - b.contact_point) <= maxd)
    {
        if (a.is_contact && b.is_contact)
            b.fake_collision = true; // Don't rebound from `b'

        DirectionBits fa = contact_faces(a);
        DirectionBits fb = contact_faces(b);

        StoneContact &c = (a.is_contact) ? a : b;
        switch (fa & fb) {
        case NORTHBIT: c.normal = V2(0,-1); break;
        case EASTBIT: c.normal = V2(1,0); break;
        case SOUTHBIT: c.normal = V2(0,1); break;
        case WESTBIT: c.normal = V2(-1,0); break;
        case NODIRBIT:
            //fprintf(stderr, "Strange: contacts have no direction in common\n");
            break;
        default:
            //fprintf(stderr, "Strange: contacts have multiple directions in common\n");
            break;
        }
    }
}

static void
find_stone_contacts (Actor *a, StoneContactList &cl)
{
    ActorInfo &ai = *a->get_actorinfo();
    px::Rect r(int(ai.pos[0]-0.5), int(ai.pos[1]-0.5),
	       int(ai.pos[0]+0.5), int(ai.pos[1]+0.5));
    r.w -= r.x;
    r.h -= r.y;

    StoneContact contacts[2][2];

    // Test for collisions with the nearby stones
    for (int i=0; i<=r.w; i++)
        for (int j=0; j<=r.h; j++)
            find_contact_with_stone(a, contacts[i][j], r.x+i, r.y+j);

    maybe_join_contacts (contacts[0][0], contacts[1][0]);
    maybe_join_contacts (contacts[0][0], contacts[0][1]);
    maybe_join_contacts (contacts[1][0], contacts[1][1]);
    maybe_join_contacts (contacts[0][1], contacts[1][1]);

    for (int i=0; i<=r.w; i++)
        for (int j=0; j<=r.h; j++)
            if (contacts[i][j].is_contact)
                cl.push_back(contacts[i][j]);
}


/* This function is called for every contact between an actor and a
   stone.  It is here that the stones are informed about contacts.
   Note that a stone may kill itself during a call to actor_hit() or
   actor_contact() so we must not refer to it after having called
   these functions.  (Incidentally, this is why StoneContact only
   contains a 'stonepos' and no 'stone' entry.)
*/
static void
handle_stone_contact (StoneContact &sc, double total_dtime)
{
    Actor *a = sc.actor;
    ActorInfo &ai = *a->get_actorinfo();

    if (sc.is_collision) {
        if (!sc.fake_collision && sc.response == STONE_REBOUND) {

            bool new_collision = false;
            if (total_dtime != 0) {
                double vn = (ai.pos - ai.last_pos)/total_dtime * sc.normal;
                new_collision = (vn < -0.5);
            }

            if (new_collision) {
                if (Stone *stone = world::GetStone(sc.stonepos)) {
                    V2 pos = ai.pos;
                    stone->actor_hit(sc);
                    if (ai.pos != pos) { // actor position was changed by actor_hit() (e.g. turnstile, pull-stone)
                        ai.oldpos = ai.pos; // HACK: avoid resetting actors position
                    }
                }
//             if (length(ai->vel) > 0.1) {
                display::AddEffect (sc.contact_point, "ring-anim");
                sound::PlaySound(sc.sound.c_str(), sc.contact_point);
//             }
            }
            double restitution = 0.95;
            ai.vel -= (1 + restitution)*(ai.vel*sc.normal)*sc.normal;
	    ai.pos = ai.oldpos;		// Reset actor position
        }
    }
    else if (sc.is_contact) {
        if (Stone *stone = world::GetStone(sc.stonepos))
            stone->actor_contact(sc.actor);
    }
}

static void
handle_contacts(Actor *actor1, double total_dtime)
{
    ActorInfo &a1 = *actor1->get_actorinfo();

    /*
    ** Handle contacts with stones
    */
    StoneContactList cl;
    find_stone_contacts(actor1, cl);
    for (StoneContactList::iterator i=cl.begin();
         i != cl.end(); ++i)
    {
        handle_stone_contact (*i, total_dtime);
    }

    /*
    ** Handle contacts with other actors
    */
    V2     p1 = a1.pos;
    double r1 = a1.radius;
    double m1 = a1.mass;

    if (!actor1->is_on_floor() || actor1->is_dead())
        return;

    unsigned s=level->actorlist.size();
    for (unsigned j=0; j<s; ++j)
    {
	Actor     *actor2 = level->actorlist[j];
        if (actor2 == actor1)
	    continue;

        if (!actor2->is_on_floor() || actor2->is_dead())
            continue;


        ActorInfo &a2 = *actor2->get_actorinfo();
        V2         p2 = a2.pos;
        double     r2 = a2.radius;

        bool contactp = length(p1-p2) < r1+r2;
        if (contactp)
        {
            V2 n  = normalize(p1-p2);	// normal to contact surface
            V2 v1 = n * (a1.vel*n);
            V2 v2 = n * (a2.vel*n);

	    a1.pos = a1.oldpos;		// Reset actor position

            bool collisionp = (v2-v1)*(p2-p1) < 0;
            if (collisionp) {
                double m2 = a2.mass;
                double M = m1+m2;

                double restitution = 0.95;

		actor1->on_hit(actor2);
		actor2->on_hit(actor1);

                a1.vel += -v1 + (v2*(2*m2) + v1*(m1-m2)) / M;
                a2.vel += -v2 + (v1*(2*m1) + v2*(m2-m1)) / M;

                a1.vel *= restitution;
                a2.vel *= restitution;
                sound::PlaySound("ballcollision");
            }
        }
    }
}

static void
tick_actor(Actor *a, double dtime)
{
    ActorInfo &ai = * a->get_actorinfo();

    ai.last_pos = ai.pos;
    ai.oldpos = ai.pos;
    ai.oldvel = ai.vel;

    double rest_time = dtime;
    double dt = 0.005; //dtime;

    handle_contacts(a, 0.0);
    while (rest_time > 0 && !ai.grabbed)
    {
        advance_actor(a, dtime-rest_time, dt);

        double l = length(ai.pos-ai.oldpos);

        // time step too large? try again
        if (l > 0.04 && dt > 1/1000.) {
            dt /= 2;
            ai.pos = ai.oldpos;
            ai.vel = ai.oldvel;
            continue;
        }

        // ok, accept this step
        a->move();
        rest_time -= dt;

        handle_contacts(a, dtime-rest_time);

        ai.oldpos  = ai.pos;
        ai.oldvel  = ai.vel;

        // time step too small?
        if (l < 0.01) {
            dt *= 2;
            dt = dt > rest_time? rest_time : dt;
        }
    }
    a->think(dtime);
}


//----------------------------------------
// Functions
//----------------------------------------

void world::Reset ()
{
    /* Restart the Lua environment so symbol definitions from
       different levels do not get in each other's way.*/
    lua::Shutdown();
    lua::Init();
    lua::Dofile("init.lua");

    enigma::GameReset();
    player::NewWorld();

    g_timer.clear();
}

void world::Create (int w, int h)
{
    level.reset (new WorldImpl(w,h));
    display::NewWorld(w, h);
}

void world::Shutdown()
{
    level.reset();
    Repos_Shutdown();
    lua::Shutdown();
}

bool world::InitWorld()
{
    level->scramble_puzzles();

    laser::RecalcLight();
    laser::RecalcLightNow();    // recalculate laser beams if necessary

    bool seen_player0 = false;

    for (ActorList::iterator i=level->actorlist.begin();
         i != level->actorlist.end(); ++i)
    {
        Actor *a = *i;
        a->on_creation(a->get_actorinfo()->pos);

        int iplayer;
        if (a->int_attrib("player", &iplayer)) {
            player::AddActor(iplayer,a);
            if (iplayer == 0) seen_player0 = true;
        }
    }

    if (!seen_player0) {
        fprintf(stderr, "Error: No player 0 defined!\n");
        return false;
    }

    world::BroadcastMessage("init", Value(),
                            // currently only EasyModeItem listens to "init" :
                            static_cast<GridLayerBits>(GRID_ITEMS_BIT));

    player::InitMoveCounter();
    display::StatusBar *sb = display::GetStatusBar();
    sb->show_move_counter (enigma::ShowMoves);

    return true;
}

void world::SetMouseForce(V2 f) {
    level->mouseforce.add_force(f);
}

void world::NameObject(Object *obj, const std::string &name) {
    string old_name;
    if (obj->string_attrib("name", &old_name)) {
        obj->warning("name '%s' overwritten by '%s'",
                     old_name.c_str(), name.c_str());
        UnnameObject(obj);
    }
    level->name_object (obj, name);
}

void world::UnnameObject(Object *obj) {
    level->unname(obj);
}

void
world::TransferObjectName(Object *source, Object *target) {
    string name;
    if (source->string_attrib("name", &name)) {
        UnnameObject(source);
        string targetName;
        if (target->string_attrib("name", &targetName)) {
            target->warning("name '%s' overwritten by '%s'",
                            targetName.c_str(), name.c_str());
            UnnameObject(target);
        }
        NameObject(target, name);
    }
}

Object * world::GetNamedObject(const std::string &name) {
    return level->get_named (name);
}

bool world::IsLevelBorder(GridPos p) {
    return level->is_border(p);
}

//----------------------------------------
// Force fields
//----------------------------------------


void
world::AddForceField(ForceField *ff)
{
    level->forces.push_back(ff);
}

void world::RemoveForceField(ForceField *ff) {
    level->remove (ff);
}

//----------------------------------------
// Rubber bands
//----------------------------------------

void world::AddRubberBand (Actor *a, Stone *st, double strength,double length)
{
    level->rubbers.push_back(new RubberBand (a, st, strength, length));
}

void world::AddRubberBand (Actor *a, Actor *a2, double strength,double length)
{
    level->rubbers.push_back(new RubberBand (a, a2, strength, length));
}

void world::KillRubberBand (Actor *a, Stone *st)
{
    assert(a);
    for (unsigned i=0; i<level->rubbers.size(); ) {
        RubberBand &r = *level->rubbers[i];
        if (r.get_actor() == a && r.get_stone() != 0)
            if (r.get_stone()==st || st==0) {
                delete &r;
                level->rubbers.erase(level->rubbers.begin()+i);
                continue;       // don't increment i
            }
        ++i;
    }
}

void world::KillRubberBand (Actor *a, Actor *a2)
{
    assert (a);
    for (unsigned i=0; i<level->rubbers.size(); ) {
        RubberBand &r = *level->rubbers[i];
        if (r.get_actor() == a && r.get_actor2() != 0)
            if (r.get_actor2()==a2 || a2==0) {
                delete &r;
                level->rubbers.erase(level->rubbers.begin()+i);
                continue;       // don't increment i
            }
        ++i;
    }
}

void world::KillRubberBands (Stone *st)
{
   for (unsigned i=0; i<level->rubbers.size(); ) {
        RubberBand &r = *level->rubbers[i];
        if (r.get_stone() != 0 && r.get_stone()==st) {
            delete &r;
            level->rubbers.erase(level->rubbers.begin()+i);
            continue;       // don't increment i
        }
        ++i;
    }
}

void
world::GiveRubberBands (Stone *st, vector<Rubber_Band_Info> &rubs)
{
   for (unsigned i=0; i<level->rubbers.size(); ) {
        RubberBand &r = *level->rubbers[i];
        if (r.get_stone() == st) {
	    Rubber_Band_Info rbi;
	    rbi.act = r.get_actor();
	    rbi.strength = r.get_strength();
	    rbi.length = r.get_length();
	    rubs.push_back(rbi);
	}
        ++i;
    }
}

bool world::HasRubberBand (Actor *a, Stone *st)
{
    for (unsigned i=0; i<level->rubbers.size(); ++i) {
        RubberBand &r = *level->rubbers[i];
        if (r.get_actor() == a && r.get_stone() == st)
            return true;
    }
    return false;
}

//----------------------------------------
// Signals
//----------------------------------------

void world::AddSignal (GridLoc srcloc, GridLoc dstloc, const string &msg)
{
#if defined(VERBOSE_MESSAGES)
    fprintf(stderr, "AddSignal src=%i/%i dest=%i/%i msg='%s'\n",
            srcloc.pos.x, srcloc.pos.y, dstloc.pos.x, dstloc.pos.y, msg.c_str());
#endif // VERBOSE_MESSAGES

    if (Object *src = GetObject(srcloc)) {
        src->set_attrib("action", "signal");
        level->signals.add (new SimpleSignal (src, dstloc, msg));
    }
    else {
        Log << "AddSignal: Invalid signal source\n";
    }
}


void world::EmitSignals (Object *src, int value)
{
    level->signals.emit_from (src, value);
}

// searches all signals for a signal starting at 'src'
// returns the target object.
// if multiple signals start from 'src' this function
// returns 0 if more than one target objects exists.
Object *world::FindSignalDestination(Object *src)
{
    return level->signals.find_single_destination(src);
}

namespace
{
//----------------------------------------
// Layer
//
// Changes to a layer are cached and only applied at the end of a
// tick. [### not implemented]
//----------------------------------------
    template <class T>
    class Layer {
        T *defaultval;
    public:
        Layer(T* deflt = 0) : defaultval(deflt) {}
        virtual ~Layer() {}

        T *get(GridPos p) {
            if (Field *f=level->get_field(p))
                return raw_get(*f);
            else
                return defaultval;
        }

        T *yield(GridPos p) {
            if (Field *f=level->get_field(p)) {
                T *x = raw_get(*f);
                if (x) {
                    raw_set(*f, 0);
                    x->removal();
                }
                return x;
            } else
                return defaultval;
        }

        void set(GridPos p, T *x) {
            if (x) {
                if (Field *f=level->get_field(p)) {
                    dispose(raw_get(*f));
                    raw_set(*f, x);
                    x->creation(p);
                }
                else
                    dispose(x);
            }
        }
        void kill(GridPos p) { dispose(yield(p)); }

        virtual T* raw_get (Field &) = 0;
        virtual void raw_set (Field &, T *) = 0;

    private:
        virtual void dispose(T *x) { if (x) DisposeObject(x); }
    };

    /*
    ** Floor layer
    */
    class FloorLayer : public Layer<Floor> {
    private:
        Floor *raw_get (Field &f) { return f.floor; }
        void raw_set (Field &f, Floor *x) { f.floor = x;}
    };


    /*
    ** Item layer
    */
    class ItemLayer : public Layer<Item> {
    private:
        Item *raw_get (Field &f) { return f.item; }
        void raw_set (Field &f, Item *x) { f.item = x;}
    };

    /*
    ** Stone layer
    */
    class StoneLayer : public Layer<Stone>
    {
    public:
        StoneLayer() : Layer<Stone>(&borderstone) {}
    private:
        Stone *raw_get (Field &f) { return f.stone; }
        void raw_set (Field &f, Stone *st) { f.stone = st;}
        void dispose (Stone *st) {
            if (st) {
                KillRubberBands(st);
                DisposeObject(st);
            }
        }

        /* This stone is used as the virtual border of the playing area.
           It is immovable and indestructible and makes sure the player's
           marble cannot leave the level. */
        class BorderStone : public Stone {
        public:
            BorderStone() : Stone("borderstone") {}
            Object *clone() { return this; }
            void dispose() {}
        };

        BorderStone borderstone;
    };

}



namespace
{
    FloorLayer      fl_layer;
    ItemLayer       it_layer;
    StoneLayer      st_layer;
    vector<GridPos> changed_stones;

    // delayed impulses :

    class DelayedImpulse {
        Impulse      impulse;
        double       delay;
        const Stone *receiver;  // to test if stone has changed

        DelayedImpulse& operator = (const DelayedImpulse& other); // forbidden
    public:
        DelayedImpulse(const Impulse& impulse_, double delay_, const Stone *receiver_)
        : impulse(impulse_), delay(delay_), receiver(receiver_) {}

        DelayedImpulse(const DelayedImpulse& other)
            : impulse(other.impulse), delay(other.delay), receiver(other.receiver) {}

        bool tick(double dtime) { // returns true if Impulse has to be sent NOW
            delay -= dtime;
            return delay <= 0;
        }

        const GridPos& destination() const { return impulse.dest; }

        void send_impulse(Stone *target) const {

            // @@@ FIXME: the test for equality of Stones sometimes fails
            // (e.g. with turnstiles rotated by rotators)
            //
            // I guess this happens, because when the turnstile-pivot removes and add arms
            // during turn, it may happen, that the Stone receives the same memory address
            //
            // Possible fix : add unique ID to all objects

            if (target == receiver) { // if object did not change since impulse was initiated
                target->on_impulse(impulse);
            }
        }
    };

    typedef list<DelayedImpulse> ImpulseList;
    ImpulseList delayed_impulses;
}


//----------------------------------------
// Floor manipulation.
//----------------------------------------

void world::KillFloor(GridPos p) {fl_layer.kill(p);}
Floor *world::GetFloor(GridPos p) {return fl_layer.get(p);}
void world::SetFloor(GridPos p, Floor* fl) {
    fl_layer.set(p,fl);
    Stone *st = GetStone(p);
    if (st)
        st->floor_change();
}

//----------------------------------------
// Stone manipulation.
//----------------------------------------

namespace
{
    void stone_change(GridPos p) {
        Stone *st = GetStone(p);
        if (st)
            st->floor_change();

        if (Item *it = GetItem(p))
            it->stone_change(st);

        if (Floor *fl = GetFloor(p))
            fl->stone_change(st);

        laser::MaybeRecalcLight(p);
    }
}

void
world::BroadcastMessage(const std::string& msg, const Value& value, GridLayerBits grids) {
    int  width     = level->w;
    int  height    = level->h;
    bool to_floors = grids & GRID_FLOOR_BIT;
    bool to_items  = grids & GRID_ITEMS_BIT;
    bool to_stones = grids & GRID_STONES_BIT;

    for (int y = 0; y<height; ++y) {
        for (int x = 0; x<width; ++x) {
            GridPos p(x, y);
            if (to_floors) if (Floor *fl = GetFloor(p)) SendMessage(fl, msg, value);
            if (to_items)  if (Item *it  = GetItem(p))  SendMessage(it, msg, value);
            if (to_stones) if (Stone *st = GetStone(p)) SendMessage(st, msg, value);
        }
    }
}

void
world::KillStone(GridPos p)
{
    st_layer.kill(p);
//    stone_change(p);
    changed_stones.push_back(p);
}

Stone *
world::GetStone(GridPos p) {return st_layer.get(p);}

Stone *
world::YieldStone(GridPos p)
{
    Stone *st = st_layer.yield(p);
    laser::MaybeRecalcLight(p);
    changed_stones.push_back(p);
    return st;
}

void
world::SetStone(GridPos p, Stone* st)
{
    st_layer.set(p,st);
    laser::MaybeRecalcLight(p);
    changed_stones.push_back(p);
}

void
world::ReplaceStone(GridPos p, Stone* st)
{
    Stone *old = st_layer.get(p);
    if (old) {
        TransferObjectName(old, st);
        st_layer.kill(p);
    }
    SetStone(p, st);
}

#if 0 // currently unused
void
world::SwapStones(GridPos p, GridPos newp)
{
    if (!IsLevelBorder(newp) && !IsLevelBorder(p)) {
        Stone *tmp  = st_layer.yield(newp);
        Stone *tmp2 = st_layer.yield(p);

        tmp->play_sound ("st-move");

        SetStone(newp, tmp2);
        SetStone(p, tmp);
        tmp->on_move();
        tmp2->on_move();

        player::IncMoveCounter(2);
    }
}
#endif

void world::MoveStone(GridPos oldPos, GridPos newPos) {
    display::Model *m = display::YieldModel(GridLoc(GRID_STONES, oldPos));
    SetStone(newPos, YieldStone(oldPos));
    display::SetModel(GridLoc(GRID_STONES, newPos), m);
}

void
world::SetScrambleIntensity(int intensity) {
    level->scrambleIntensity = intensity;
}

void
world::AddScramble(GridPos p, Direction d) {
    level->add_scramble(p, d);
}


//----------------------------------------
// Item manipulation.
//----------------------------------------

void world::KillItem(GridPos p) {
    laser::MaybeRecalcLight(p);
    it_layer.kill(p);
}
Item *world::GetItem(GridPos p) {return it_layer.get(p);}
Item *world::YieldItem(GridPos p) {
    laser::MaybeRecalcLight(p);
    return it_layer.yield(p);
}
void world::SetItem(GridPos p, Item* it) {
    laser::MaybeRecalcLight(p);
    it_layer.set(p,it);
}

void world::SetItemAsStone(GridPos p, Item* it) {
    Stone *st = world::ConvertToStone(it);
    if (st)
        SetStone(p, st);
    else
        fprintf(stderr, "Could not SetItemAsStone(%s)\n", it->get_kind());
}

// ----------------------------------
//      Item <-> Stone conversion
// ----------------------------------

// These conversions assume that the name of the item starts with
// 'it-' and the name of stone starts with 'st-'. The rest of
// the name has to be identical.

// If the conversion fails, a message is printed to stderr
// and the source (item or stone) is not disposed.

Stone *world::ConvertToStone(Item *it) {
    string kind = it->get_kind();
    if (kind[0] == 'i') kind[0] = 's';

    Stone *st = MakeStone(kind.c_str());
    if (st) it->dispose();
    return st;
}
Item *world::ConvertToItem(Stone *st) {
    string kind = st->get_kind();
    if (kind[0] == 's') kind[0] = 'i';

    Item *it = MakeItem(kind.c_str());
    if (it) st->dispose();

    return it;
}

//----------------------------------------
// Actor manipulation.
//----------------------------------------

void
world::AddActor(double x, double y, Actor* a)
{
    V2 pos(x,y);
    level->actorlist.push_back(a);
    a->get_actorinfo()->pos = pos;

    // now done in AddActor(int, Actor*) :
    //     a->on_creation(pos);
    //     ReleaseActor(a);
}

Actor *
world::YieldActor(Actor *a)
{
    ActorList::iterator i=find(level->actorlist.begin(), level->actorlist.end(), a);
    if (i!=level->actorlist.end()) {
        level->actorlist.erase(i);
        GrabActor(a);
        return a;
    }
    return 0;
}

void world::WarpActor(Actor *a, double newx, double newy, bool fast)
{
    V2 newpos = V2(newx, newy);
    a->get_actorinfo()->vel = V2();
    a->warp(newpos);
}

void world::RespawnActor(Actor *a) {
    a->respawn();
}

void world::GrabActor(Actor *a)
{
    a->get_actorinfo()->grabbed = true;
}

void world::ReleaseActor(Actor *a)
{
    a->get_actorinfo()->grabbed = false;
}

bool world::GetActorsInRange (px::V2 center, double range,
                              vector<Actor*>& actors)
{
    ActorList &al = level->actorlist;
    for (ActorList::iterator i=al.begin(); i!=al.end(); ++i) {
        Actor *a = *i;
        if (length(a->get_pos()-center) < range)
            actors.push_back(a);
    }
    return !actors.empty();
}

//----------------------------------------
// handling of delayed impulses
//----------------------------------------

void
world::addDelayedImpulse(const Impulse& impulse, double delay, const Stone *estimated_receiver) {

    // @@@ FIXME: is a special handling necessary if several impulses hit same destination ?

    delayed_impulses.push_back(DelayedImpulse(impulse, delay, estimated_receiver));
}


//----------------------------------------
// Tick
//----------------------------------------

void
world::Tick(double dtime)
{
    const double timestep = 15.0/1000;
    static double timeaccu = 0.0;

    timeaccu += dtime;

    while (timeaccu >= timestep)
    {
	// Move actors
        for (ActorList::iterator i=level->actorlist.begin(); i!=level->actorlist.end(); ++i) {
            ActorInfo *ai = (*i)->get_actorinfo();
            if (!ai->grabbed && !(*i)->is_dead())
            {
		ai->force= ai->forceacc;
		ai->forceacc = V2();
                tick_actor(*i, timestep);
            }
        }

        // Handle delayed impulses
        for (ImpulseList::iterator i = delayed_impulses.begin(); i != delayed_impulses.end(); ) {
            // should the impulse take effect now ?
            if (i->tick(timestep)) {
                // is there a stone on the destination field ?
                if (Stone *st = GetStone(i->destination())) {
                    i->send_impulse(st);
                }
                i = delayed_impulses.erase(i);
            }
            else
                ++i;
        }

        // Tell floors and items about new stones.
        for (unsigned i=0; i<changed_stones.size(); ++i)
            stone_change(changed_stones[i]);
        changed_stones.clear();

        level->tick(timestep);

        g_timer.tick(timestep);
        laser::RecalcLightNow();   // recalculate laser beams if necessary
        timeaccu -= timestep;
        timeaccu = 0;
    }
}

void
world::Init()
{
    actors::Init();
    laser::Init();
    items::Init();
    stones::Init();
}
